%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   This script implements low rank light field factorization as described
%   in the paper. The operations are equivalent to the conventional NMF 
%   update rules. 
%
%   Although this Matlab script it is slower than the conventional NMF update 
%   rules, the projection operations directly map to hardware-acclerated 
%   OpenGL commands that make it magnitues faster on the GPU.
%
%   Gordon Wetzstein
%   Research Scientist
%   MIT Media Lab
%   http://media.mit.edu/~gordonw
%   April 2014
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clear all; clf;

%smallNumber = eps;
smallNumber = 1e-6;

% show the factored images during optimization
bPlotFactors    = true;

% number of time-multiplexed frames
numFrames       = 20;

% scale for intensity
beta            = 1;

% number of iterations 
numIters        = 75;

% load 2D light field
L = mean(im2double(imread('LightField.png')),3);
lightFieldResolution = size(L);

% initialize g and h with random values
G = rand([lightFieldResolution(1) numFrames]);
H = rand([lightFieldResolution(2) numFrames]);

% create matrices
PHI = repmat(eye(lightFieldResolution(1)), [lightFieldResolution(2) 1]);
PHI = sparse(PHI);

PSI = zeros([lightFieldResolution(1)*lightFieldResolution(2) lightFieldResolution(2)]);
T   = zeros([lightFieldResolution(1) lightFieldResolution(2)]);
T(:,1) = 1;
for k=1:size(L,2)
    PSI( (k-1)*lightFieldResolution(1)+1:k*lightFieldResolution(1),:) = circshift(T, [0 k-1]);
end
PSI=sparse(PSI);

% convergence
PSNRS = zeros([1 numIters]);

% for plotting
numRows = 1; numCols = 2;
if bPlotFactors
    numRows = 3;
end


for k=1:numIters
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % compute current estimate 
    L_tilde = zeros([prod(lightFieldResolution) 1]);
    for kf=1:numFrames
        L_tilde = L_tilde + ( (PHI*G(:,kf)) .* (PSI*H(:,kf)) );
    end  
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % 1. update G
                    
    for kf=1:numFrames
                                   
        % accumulated estimation with current layer set to 1
        PH = zeros([prod(lightFieldResolution) 1]);

        % this is tested to be equivalent to NMF and converges well
        PH  = PH + (PSI*H(:,kf));

        G(:,kf) = G(:,kf) .* ( PHI'*(PH.*L(:)) ) ./ ( ( PHI'*(PH.*L_tilde) ) + smallNumber); 
       
        % clamp to feasible range
        G(G<0)=0; 
        G(G>1)=1;
    end    
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % compute current estimate 
    L_tilde = zeros([prod(lightFieldResolution) 1]);
    for kf=1:numFrames
        L_tilde = L_tilde + ( (PHI*G(:,kf)) .* (PSI*H(:,kf)) );
    end  
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % 2. update H
            
    for kf=1:numFrames
        
        % accumulated estimation without current frame
        PG = zeros([prod(lightFieldResolution) 1]);
        
        % this is tested to be equivalent to NMF and converges well
        PG = PG + PHI*G(:,kf);            
        
        H(:,kf) = H(:,kf) .* ( PSI'*(PG.*L(:)) ) ./ ( ( PSI'*(PG.*L_tilde) ) + smallNumber); 
        
        % clamp to feasible range
        H(H<0)=0; 
        H(H>1)=1;
    end
        
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % plot intermediate results
    
    subplot(numRows,numCols,1); 
    imshow(L);
    title(['Target 2D Light Field (Rank-' num2str(min(size(L))) ')']);
    
    % current estimate
    L_tilde = zeros([prod(lightFieldResolution) 1]);
    for kf=1:numFrames
        L_tilde = L_tilde + ( (PHI*G(:,kf)) .* (PSI*H(:,kf)) );
    end
        
    
    MSE     = mean( (L_tilde-L(:)).^2 );
    PSNR    = 10 * log10( 1 / MSE );
    PSNRS(k) = PSNR;
    
    subplot(numRows,numCols,2); 
    imshow( reshape(L_tilde, lightFieldResolution));
    title(['Rank-' num2str(numFrames) ' Approximation, iter ' num2str(k) ', PSNR ' num2str(PSNR) ' dB']);
    
    if bPlotFactors
        subplot(numRows,numCols,3); 
        imagesc(G);
        title('G');
        
        subplot(numRows,numCols,4); 
        imagesc(H);
        title('H');
        
        subplot(numRows,numCols,[5 6]); 
        plot(PSNRS);
        xlabel('Iteration');
        ylabel('PSNR');
        title('Convergence');
    end
    drawnow;
end



