%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   This script loads an image and factors it into a time-multiplexed
%   sequence of frames for two SLMs (one in-focus and the other
%   out-of-focus). The target image has a higher contrast and resolution
%   than both SLMs (super-resolution is modeled by a Gaussian convolution
%   of both SLMs).
%
%   This script demonstrates both superresolution and also increased
%   contrast.
%
%   In practice, the PSFs for the display prototype are measured.
%
%   Gordon Wetzstein
%   Research Scientist
%   MIT Media Lab
%   http://media.mit.edu/~gordonw
%   April 2014
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clear all; clf;

% show the factored images during optimization
bPlotFactors    = true;

% number of time-multiplexed frames
numFrames       = 2;

% scale for intensity
beta            = 1;

% number of iterations 
numIters        = 50;

% blacklevel of both SLMs
black           = 0.25;

% target image - just grayscale
I = beta * mean(im2double(imread('BigBuckBunny.png')),3);

% convolution kernel for front (in-focus) layer 
kg = fspecial('gaussian', [20 20], 2);  % convolution by this factor models the super-resolved image formation
% convolution kernel for rear (out-of-focus) layer
kh = fspecial('gaussian', [20 20], 5);  % convolution by this factor models the super-resolved image formation as well as optical defocus

% initialize g and h
G = rand([size(I,1) size(I,2) numFrames]);
H = rand([size(I,1) size(I,2) numFrames]);

% save PSNR history
PSNRS = zeros([1 numIters]);

if bPlotFactors
    numRows = 1+numFrames;
else
    numRows = 1;
end
    
for k_it=1:numIters
    
    % for each layer
    for k_layer = 1:2          
        
        % compute current estimate with all layers
        I_tilde = zeros(size(I));
        for kk_frame=1:numFrames
            I_tilde = I_tilde + imfilter(H(:,:,kk_frame),kh, 0) .* imfilter(G(:,:,kk_frame),kg, 0);
        end
           
        % for each frame
        for k_frame=1:numFrames                                
            
            % compute current estimate without current layer 
            if k_layer == 1
                I_bar = imfilter(H(:,:,k_frame),kh, 0);
            else
                I_bar = imfilter(G(:,:,k_frame),kg, 0);
            end             

            % update layer
            if k_layer == 1
                G(:,:,k_frame) = G(:,:,k_frame) .* imfilter(I_bar.*I,kg, 0) ./ imfilter(I_bar.*I_tilde,kg, 0);
                G(G<black)=black; G(G>1)=1;
            else
                H(:,:,k_frame) = H(:,:,k_frame) .* imfilter(I_bar.*I,kh, 0) ./ imfilter(I_bar.*I_tilde,kh, 0);
                H(H<black)=black; H(H>1)=1;
            end
            
        end

    end
    
    estimate = zeros(size(I));
    for kkk=1:numFrames
        estimate = estimate + imfilter(H(:,:,kkk),kh, 0) .* imfilter(G(:,:,kkk),kg, 0);
    end
    MSE         = mean( (I(:)-estimate(:)).^2);
    PSNR        = 10 * log10(beta^2/MSE);
    PSNRS(k_it) = PSNR;
    
    subplot(numRows,3,1); imshow(I/beta); title('Target Image');
    subplot(numRows,3,2); imshow(estimate/beta); title('current');
    title(['Super-resolved (proposed), iter ' num2str(k_it) ' PSNR: ' num2str(PSNR)]);
        
    reference = imfilter(I,kg)/beta;
    reference(reference<black) = black;
    subplot(numRows,3,3); imshow(reference); title('In-focus SLM only (conventional)');
    
    
    if bPlotFactors
        for kkk=1:numFrames
            subplot(numRows,3,kkk*3+1); imshow(G(:,:,kkk)); title(['G, frame ' num2str(kkk)]);
            subplot(numRows,3,kkk*3+2); imshow(H(:,:,kkk)); title(['H, frame ' num2str(kkk)]);
        end
    
        subplot(numRows,3,6); 
        plot(PSNRS);
        ylim([25 30]);
        xlabel('Iterations');
        ylabel('PSNR');
        title('Convergence');
    end
    
    drawnow;
end
